
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "rhp.h"
#include "sif_f.h"
#include "cell.h"
#include "sif.h"

extern void	Error(char *msg);

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)(A),0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)(A))
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

#define INIT_MT(T,F,A)	\
if (INIT_THREAD(T,F,A)) Error(#F " : multi-threading initialization error.")
#define TERM_MT(T,F)	\
if (TERM_THREAD(T)) Error(#F " : multi-threading termination error.")

#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif

#ifndef	Float
#define Float	float
#endif

#ifndef	Filter
#define Filter	Chesler
#endif

double	Ramachandran(int j)
{
	return (j==0)?1.0/4.0:(j&1)?-1.0/(M_PI*M_PI*(double)j*(double)j):0.0;
}

double	Shepp(int j)
{
	return 2.0/(M_PI*M_PI*(1.0-4.0*(double)j*(double)j));
}

double	Chesler(int j)
{
	if (j==0)
	    return 1.0/8.0-1.0/(2.0*M_PI*M_PI);
	else if (j==1 || j==(-1))
	    return 1.0/16.0-1.0/(2.0*M_PI*M_PI);
	else if (j&1)
	    return -1.0/(2.0*M_PI*M_PI*(double)j*(double)j);
{
	double  j2=(double)j*(double)j,j21=j2-1.0;

	return -(j2+1.0)/(2.0*M_PI*M_PI*j21*j21);
}}

extern double	Filter(int j);

typedef struct {
		double	x,y;
	} Vector;

typedef struct {
		int	j;
		double	r;
	} Struct;

static int	Nr,N1,N2,N,L,Threads=1;
static Vector	*E,*G,**W;
static double	**P,**Q,f4cbp;
static Struct	**S4BP;
static Float	***F;

static void	FFT(int sign,Vector *V)
{
	int	j,k,l,m,p,q;
	Vector	v,*e,*a,*b;
	double	x,y,z;

	for (j=N2-1, k=N2-2; k>0; k--) {
	    for (l=N1; (j^=l)&l; l>>=1) ;

	    if (j>k) {
		v=V[j]; V[j]=V[k]; V[k]=v;
	    }
	}
	for (j=2, k=1, l=N1, m=N2; m>1; m=l, l>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=m) {
		x=e->x; y=e->y*(double)sign;
		for (a=V+p, q=0; q<l; q++, a+=j) {
		    v=(*(b=a+k)); b->x=a->x-(z=v.x*x-v.y*y); a->x+=z;
				  b->y=a->y-(z=v.x*y+v.y*x); a->y+=z;
		}
	    }
}

static FUNCTION_T	Convolution(void *i)
{
	size_t	l;
	Vector	*v,*g,w,*V=W[(size_t)i];
	double	*p;
	int	j;

	for (l=(size_t)i; l<L; l+=Threads) {
	    for (v=V, j=0; j<N1; j++, v++) v->x=v->y=0.0;

	    p=P[l];
	    for (j=0; j<Nr; j++, v++) {
		v->x=(*(p++)); v->y=0.0;
	    }
	    for (; j<N1; j++, v++) v->x=v->y=0.0;

	    FFT(-1,V);

	    for (v=V, g=G, j=0; j<N2; j++, v++, g++) {
		w=(*v); v->x=g->x*w.x-g->y*w.y;
			v->y=g->x*w.y+g->y*w.x;
	    }
	    FFT(1,V);

	    for (v=V, j=0; j<Nr; j++, v++) Q[j][l]=f4cbp*v->x;
	}
	return RETURN_VALUE;
}

static FUNCTION_T	BackProjection(void *i)
{
	size_t	y;
	int	x,j,z;
	Float	**Fy,*Fyx;
	Struct	*s4bp;
	double	*Q0,*Q1,r0,r1;

	for (y=(size_t)i; y<N; y+=Threads) {
	    Fy=F[y]; s4bp=S4BP[y];
	    for (x=0; x<N; x++)
		if ((j=s4bp[x].j)>=0) {
		    Fyx=Fy[x]; Q0=Q[j]; Q1=Q[j+1]; r0=1.0-(r1=s4bp[x].r);
		    for (z=0; z<L; z++) *(Fyx++)+=(*(Q0++)*r0+(*(Q1++))*r1);
		}
	}
	return RETURN_VALUE;
}

#define LEN	2048
#define EPS	1e-9

int	main(int argc,char **argv)
{
	char		*env,path[LEN];
	HiPic		hp;
	int		Nl,M,O,l1,l2,r,z,Y,y,X,x,m,l,bps=0;
	double		Dr,Or,ra0,*p,S,C,R;
	THREAD_T	*Thread;
	size_t		i;

	if (argc!=6 && argc!=7 && argc!=8 && argc!=9) Error(
	    "usage : hp_cbp HiPic/ Dr Or {layer1 layer2} RA0 {BPS} TG_format"
	);
	if ((env=getenv("THREADS"))!=NULL &&
	    (Threads=atoi(env))<=0) Error("bad number of THREADS.");

	InitReadHiPic(argv[1],&hp); Nr=hp.Nx; Nl=hp.Ny; M=hp.Nt;

	if ((Dr=atof(argv[2]))<EPS) Error("bad interval of detectors.");

	if ((Or=atof(argv[3]))<=0.0 || Or>=(double)(Nr-1))
	    Error("bad rotation center (used as origin of detector array).");

	for (N1=1; N1<Nr; N1<<=1) ; N2=N1<<1;

	N=(O=(int)floor(((double)(Nr-1)/2.0<Or)?(double)(Nr-1)-Or:Or))*2+1;

	if (argc==6 || argc==7) {
	    l1=0; l2=Nl-1;
	}
	else
	    if ((l1=atoi(argv[4]))<0 ||
		(l2=atoi(argv[5]))>=Nl || l1>l2) Error("bad layer range.");

	L=l2-l1+1;

	if (argc!=6 && argc!=8 && abs(bps=atoi(argv[argc-2]))>sizeof(Cell)*8)
	    Error("bad bits per sample (BPS).");

	(void)fprintf(stderr,"%d\t%d\t%s\n",N,L,argv[2]);

	if ((Thread=(THREAD_T *)malloc(sizeof(THREAD_T)*Threads))==NULL ||
	    (E=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (G=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (W=(Vector **)malloc(sizeof(Vector *)*Threads))==NULL ||
	    (*W=(Vector *)malloc(sizeof(Vector)*Threads*N2))==NULL ||
	    (P=(double **)malloc(sizeof(double *)*L))==NULL ||
	    (*P=(double *)malloc(sizeof(double)*L*Nr))==NULL ||
	    (Q=(double **)malloc(sizeof(double *)*(Nr+1)))==NULL ||
	    (*Q=(double *)malloc(sizeof(double)*Nr*L))==NULL ||
	    (S4BP=(Struct **)malloc(sizeof(Struct *)*N))==NULL ||
	    (*S4BP=(Struct *)malloc(sizeof(Struct)*N*N))==NULL ||
	    (F=(Float ***)malloc(sizeof(Float **)*N))==NULL ||
	    (*F=(Float **)malloc(sizeof(Float *)*N*N))==NULL ||
	    (**F=(Float *)malloc(sizeof(Float)*N*N*L))==NULL)
	    Error("memory allocation error.");

	for (r=0; r<N2; r++) {
	    sincos(M_PI*(double)r/(double)N2,&(E[r].y),&(E[r].x));

	    G[r].x=Filter(r-N1); G[r].y=0.0;
	}
	FFT(-1,G);

	for (i=1; i<Threads; i++) W[i]=W[i-1]+N2;

	for (z=1; z<L; z++) P[z]=P[z-1]+Nr;

	Q[Nr]=Q[0]; for (r=1; r<Nr; r++) Q[r]=Q[r-1]+L;
{
	int	O2_Y2,O2=O*O;

	for (Y=O, y=0; y<N; y++, Y--) {
	    if (y>0) {
		S4BP[y]=S4BP[y-1]+N; F[y]=F[y-1]+N;
	    }
	    O2_Y2=O2-Y*Y;
	    for (X=(-O), x=0; x<N; x++, X++) {
		S4BP[y][x].j=(O2_Y2>=X*X)-1;

		if (y>0 || x>0) F[y][x]=F[y][x-1]+L;

		for (z=0; z<L; z++) F[y][x][z]=0.0;
	    }
	}
}
	ra0=atof(argv[(argc==6 || argc==8)?argc-2:argc-3])*M_PI/180.0;

	f4cbp=1.0/Dr/(double)N2*M_PI/(double)M;

	for (m=0; m<M; m++) {
	    ReadHiPic(&hp,m);

	    for (l=l1; l<=l2; l++) {
		p=P[l-l1];
		for (r=0; r<Nr; r++)
		    *(p++)=(hp.T[l][r]>0.0)?-log(hp.T[l][r]):0.0;
	    }
	    if (m>0)
		for (i=0; i<Threads; i++) TERM_MT(Thread[i],BackProjection);

	    for (i=0; i<Threads; i++) INIT_MT(Thread[i],Convolution,i);

	    sincos(ra0+M_PI*(double)m/(double)M,&S,&C);
	    for (Y=  O , y=0; y<N; y++, Y--)
	    for (X=(-O), x=0; x<N; x++, X++)
		if (S4BP[y][x].j>=0) {
		    S4BP[y][x].j=r=(int)(R=Or+C*(double)X+S*(double)Y);
		    S4BP[y][x].r=R-(double)r;
		}

	    for (i=0; i<Threads; i++) TERM_MT(Thread[i],Convolution);
	    for (i=0; i<Threads; i++) INIT_MT(Thread[i],BackProjection,i);
	}
	for (i=0; i<Threads; i++) TERM_MT(Thread[i],BackProjection);

	free(*Q); free(Q); free(*P); free(P); free(*W); free(W);
	free(G); free(E); free(Thread);

	TermReadHiPic(&hp);

	if (bps==0)
{
	FOM	**fom=(FOM **)S4BP;

	for (z=0; z<L; z++) {
	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++) fom[y][x]=F[y][x][z];

	    (void)sprintf(path,argv[argc-1],z+l1);
	    StoreImageFile_Float(path,N,N,fom,SIF_F_desc);

	    (void)printf("%d\t%s\n",z+l1,SIF_F_desc);
	}
}
	else
{
	Cell	**cell=(Cell **)S4BP;
	double	min,max,base,step;
	char	desc[LEN];

	if (bps<0) {
	    min=max=F[0][0][0];
	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++)
	    for (z=0; z<L; z++) if (min>F[y][x][z]) min=F[y][x][z]; else
				if (max<F[y][x][z]) max=F[y][x][z];

	    base=min;
	    step=(max-min<EPS)?1.0:(max-min)/(double)((1ULL<<(-bps))-1);
	    (void)sprintf(desc,SIF_F_FORM,min,max);

	    (void)fprintf(stderr,"%s\n",desc);
	}
	for (z=0; z<L; z++) {
	    min=max=F[0][0][z];
	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++) if (min>F[y][x][z]) min=F[y][x][z]; else
				if (max<F[y][x][z]) max=F[y][x][z];

	    if (bps>0) {
		base=min;
		step=(max-min<EPS)?1.0:(max-min)/(double)((1ULL<<bps)-1);
		(void)sprintf(desc,SIF_F_FORM,min,max);
	    }
	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++) cell[y][x]=(Cell)((F[y][x][z]-base)/step+0.5);

	    (void)sprintf(path,argv[argc-1],z+l1);
	    StoreImageFile(path,N,N,abs(bps),cell,desc);

	    (void)printf("%d\t" SIF_F_FORM "\n",z+l1,min,max);
	}
}
	free(**F); free(*F); free(F); free(*S4BP); free(S4BP);

	return 0;
}
